#include <cmath>
#include <queue>
#include <stack>
#include "leveldb/db.h"
#include <openssl/aes.h>

using namespace std;

unsigned char aeskey[32] = {0};
uint64_t total_time = 0;

#define FP_SIZE 6
#define MAX_UO 4000 // MAX left/right entropy of M 4000
#define MAX_UT 4000 // MAX left/right entropy of C 4000

#define W 200000 // maxinum of entropy

#define MAX_INT 9999999 
#define E_ATTACK 1 // Encryption
#define MIN_ENTROPY 0

#define LEFT_NEIGHBOR_FLAG 101
#define RIGHT_NEIGHBOR_FLAG 202


// TOP_FREQUENT_U:
//    - Apply analysis to infer top-frequent(aka, u) ciphertext-plaintext pairs {(C,M)} from C and M
//    - Indicate the number of most frequent chunk pairs to be returned
uint64_t TOP_FREQUENT_U;

// r->ridus
int RADIUS; 

// t->threshold
double ENTROPY_THRESHOLD; 

uint64_t NEIGHBOR_TOP_FREQUENT_V; 
uint64_t QUEUE_LIMIT_W;
double LEAKAGE_RATE;

uint64_t smallChunkAmount = 0; 
uint64_t bigChunkError = 0; 
uint64_t smallChunkError = 0;

// chunk_node: key->chunkhash, count->frequency/co-occurrence frequency
struct node 
{
    char key[FP_SIZE];
    uint64_t count;
    uint64_t total_time;
};

/* comparison function of priority queue, the smaller count is at the head of queue, deleted first */
struct cmp 
{
    bool operator()(node a, node b)
    {
        return a.count > b.count;
    }
};

leveldb::DB *origin_M; // F_M, F_Plaintext
leveldb::DB *left_M; // L_M, Left_Plaintext
leveldb::DB *right_M; // R_M, Right_Plaintext

leveldb::DB *target_C; // F_C, F_Ciphertext
leveldb::DB *left_C; // L_C, Left_Ciphertext
leveldb::DB *right_C; // R_C, Right_Ciphertext

// unique db is used to record all inferred chunks (correct & error)
leveldb::DB *inferred_chunks_DB; // store all unique for inferring chunks

// refer to ground-truth db, using for simulating attack
leveldb::DB *ground_truth_DB; 


// temp priority_queue sorted by frequency(top u frequency & co-occurrence frequency chunk node of M/C)
priority_queue<node, vector<node>, cmp> frequency_queue_M;
priority_queue<node, vector<node>, cmp> frequency_queue_C;

// queue_M_G and queue_C_G implement inferred set G
// one-to-one correspondence between M and C in these two queues - inferred pair
queue<node> queue_M_G; // store all ciphertext in G, also means M in G
queue<node> queue_C_G; // store all plaintext in G, also means C in G

// unique chunk size
uint64_t total = 0; 

uint64_t correct = 0; 
uint64_t involve = 0; 

uint64_t leaked_common = 0; 
uint64_t leaked_chunks = 0; 

// correctly inferred cunk
vector<node> ansq;

/* init db */
void init_db(char *db_name, leveldb::DB **db)
{
    leveldb::Options options;
    options.create_if_missing = true;
    leveldb::Status status = leveldb::DB::Open(options, db_name, db);
    if (!status.ok()) {
        printf("[Error] Failed to open DB!\n");
        exit(-1);
    }
    if (db == nullptr) {
        printf("[Error] Failed to assign DB pointer!\n");
        exit(-1);
    }
}

/* enqueue leaked chunks based on leakage rate */
void enqueue_leaked_chunks()
{
    leveldb::Status status;
    std::string existing_value;
    leveldb::Iterator *it = target_C->NewIterator(leveldb::ReadOptions());

    if (LEAKAGE_RATE == 0.00) { 
        for (it->SeekToFirst(); it->Valid(); it->Next()) {
            total++;
        }
        printf("Total number of unique ciphertext chunks: %ld\nZero Leakage rate: %lf%%\n", total,
              (double)(LEAKAGE_RATE * 100.0));
        delete (it); 
        return;
    }

    for (it->SeekToFirst(); it->Valid(); it->Next()) {
        total++;
        status = origin_M->Get(leveldb::ReadOptions(), it->key(), &existing_value);
        if (rand() % 10000 <= LEAKAGE_RATE * 10000) {
            leaked_common++;
            if (status.ok()) {
                node node_M{};
                memcpy(node_M.key, it->key().ToString().c_str(), FP_SIZE);
                node_M.count = *(uint64_t *)existing_value.c_str();
                queue_M_G.push(node_M);

                node node_C{};
                memcpy(node_C.key, it->key().ToString().c_str(), FP_SIZE);
                node_C.count = *(uint64_t *)it->value().ToString().c_str();

                leveldb::Status s;
                leveldb::Slice key(node_C.key, FP_SIZE);
                char buf[32];
                memset(buf, 0, 32);
                sprintf(buf, "%ld", node_C.count);
                leveldb::Slice value(buf, sizeof(uint64_t));
                s = inferred_chunks_DB->Put(leveldb::WriteOptions(), key, value);
                queue_C_G.push(node_C);

                leaked_chunks++;
            } else if (!status.IsNotFound())
                printf("%s\n", status.ToString().c_str());
        }
    }
    delete it; 

    printf("Total number of unique ciphertext chunks: %ld\nLeakage rate: %lf%%\n", total,
           (double)(LEAKAGE_RATE * 100.0));
}

/* clear data in priority_queue */
void clear(priority_queue<node, vector<node>, cmp> &queue) {
    while (!queue.empty()) {
        queue.pop();
    }
}

/* clear data in stack */
void clear(stack<node> &stack_node) {
    while (!stack_node.empty()) {
        stack_node.pop();
    }
}

/* print chunkhash of a node */
void print_fp(node a)
{
    int len = 0;
    for (len = 0; len < FP_SIZE; len++) {
        printf("%02x", (unsigned char)a.key[len]);
        if (len < FP_SIZE - 1) printf(":");
    }
    printf("\n");
}

void print_fp(char * a)
{
    int len = 0;
    for (len = 0; len < FP_SIZE; len++) {
        printf("%02x", (unsigned char)a[len]);
        if (len < FP_SIZE - 1) printf(":");
    }
    printf("\n");
}

/* insert top-k neighbor_chunks with the highest co-occurrence frequency (in neighbor_db) into frequency_queue_M/frequency_queue_C */
void neighbor_DB_insert(int neighbor_flag, int type, char *fp, uint64_t k)
{
    leveldb::Status status;
    leveldb::Slice key(fp, FP_SIZE);
    std::string existing_value;
    priority_queue<node, vector<node>, cmp> *pt;    
    if (type == 0)
        pt = &frequency_queue_M;
    else if (type == 1)
        pt = &frequency_queue_C;

    leveldb::DB *neighbor_M_DB;
    leveldb::DB *neighbor_C_DB;

    if (neighbor_flag == LEFT_NEIGHBOR_FLAG) {
        neighbor_M_DB = left_M;
        neighbor_C_DB = left_C;
    } else if (neighbor_flag == RIGHT_NEIGHBOR_FLAG) {
        neighbor_M_DB = right_M;
        neighbor_C_DB = right_C;
    } else {
        printf("[Error] Not right neighbor_flag chosen");
        exit(-1);
    }

    uint64_t len = 0;
    char tar[FP_SIZE];
    uint64_t tmp;
    
    if (type == 0)
        status = neighbor_M_DB->Get(leveldb::ReadOptions(), key, &existing_value);
    else
        status = neighbor_C_DB->Get(leveldb::ReadOptions(), key, &existing_value);

    if (status.ok()) {

        while (len < existing_value.size() - 1) {

            memcpy(tar, existing_value.c_str() + len, FP_SIZE);
            const char *t_int = existing_value.c_str() + len + FP_SIZE;
            tmp = *(uint64_t *)t_int;

            node entry;
            memcpy(entry.key, existing_value.c_str() + len, FP_SIZE);
            entry.count = tmp;
            entry.total_time = total_time++;

            if (pt->size() < k) {
                pt->push(entry);
            } else {
                node min = pt->top();
                if (tmp > min.count) {
                    pt->pop();
                    pt->push(entry);
                }
            }

            len += (FP_SIZE + sizeof(uint64_t));
        }
    } else if (!status.IsNotFound())
        printf("%s\n", status.ToString().c_str());
}

/* insert top-k (parameter u in paper) frequent chunks (in db) into frequency_queue_M */
void db_insert(leveldb::DB *db, uint64_t k)
{
    leveldb::Iterator *it = db->NewIterator(leveldb::ReadOptions());
    for (it->SeekToFirst(); it->Valid(); it->Next()) {
        uint64_t val = *(uint64_t *)it->value().ToString().c_str();
        node entry;
        memcpy(entry.key, it->key().ToString().c_str(), FP_SIZE);
        entry.count = val;
        entry.total_time = total_time++;

        if (frequency_queue_M.size() < k) {
            frequency_queue_M.push(entry);
        } else {
            node min = frequency_queue_M.top();
            if (frequency_queue_M.top().count < val) {
                // push chunks with more frequency
                frequency_queue_M.pop();
                frequency_queue_M.push(entry);
            }
        }
    }
    delete it;

}

/* calculate chunk's entropy */
double entropy(vector<node> &tp)
{
    double ans_entropy = 0;
    uint64_t count_sum = 0;
    double min_entropy = 9999999;
    vector<node>::iterator it;

    for (it = tp.begin(); it != tp.end(); it++) {
        count_sum += it->count;
    }
    if (count_sum == 0) return 0;

    for (it = tp.begin(); it != tp.end(); it++) {
        double po = (double)it->count / (double)count_sum;
        //printf("po: %llu %llu %f \n", it->count, count_sum,po);
        if (po != 0) {
            ans_entropy += (-po * (log(po) / log(2)));
            if ((-po * (log(po) / log(2)) < min_entropy)) min_entropy = (-po * (log(po) / log(2)));
        }
    }

    if (MIN_ENTROPY) ans_entropy = min_entropy;
    if(ans_entropy > 9999) printf("entropy: %f\n", ans_entropy);
    return ans_entropy;
}

/* get all the chosen chunks' entropy */
void get_entropy(stack<node> &in_stack, double *left_entropies, double *right_entropies, int type, int num)
{
    double left_entropy, right_entropy;
    // rename
    vector<node> cal;
    stack<node> inverted;
    priority_queue<node, vector<node>, cmp> *ui;

    if (type == 0) {
        ui = &frequency_queue_M;
    } else {
        ui = &frequency_queue_C;
    }

    int ans_position = 0;
    // initialize internal_stack with in_stack (copy only)
    
    stack<node> internal_stack(in_stack);

    while (!internal_stack.empty()) {
        node top_node = internal_stack.top();
        internal_stack.pop();

        cal.clear();
        clear(inverted);
        clear(frequency_queue_M);
        clear(frequency_queue_C);
        neighbor_DB_insert(LEFT_NEIGHBOR_FLAG, type, top_node.key, num);
        
        while (!ui->empty()) {
            inverted.push(ui->top());
            ui->pop();
        }

        while (!inverted.empty()) {
            cal.push_back(inverted.top());
            inverted.pop();
        }

        left_entropy = entropy(cal);
        // printf("%lf\n",left_entropy);
        //--------------------------------------------------
        cal.clear();
        clear(inverted);
        clear(frequency_queue_M);
        clear(frequency_queue_C);

        neighbor_DB_insert(RIGHT_NEIGHBOR_FLAG, type, top_node.key, num);

        while (!ui->empty()) {
            inverted.push(ui->top());
            ui->pop();
        }
        while (!inverted.empty()) {
            cal.push_back(inverted.top());
            inverted.pop();
        }
        right_entropy = entropy(cal);
        //----------------------------------------------------
        left_entropies[ans_position] = left_entropy;
        right_entropies[ans_position] = right_entropy;
        ans_position++;
    }
}

/* compute euclidean distance -- d(Ci,Mj) */
double dist(double l1, double r1, double l2, double r2)
{
    return sqrt((l1 - l2) * (l1 - l2) + (r1 - r2) * (r1 - r2));
}

/* infer plaintext(M)-ciphertext(C) pair according to euclidean distance */
void peakToPair(double leftOrigin[], double rightOrigin[], double leftTarget[], double rightTarget[],
                stack<node> &originStack, stack<node> &targetStack,int flag)
{
    vector<node> originTemp, targetTemp;
    originTemp.clear();
    targetTemp.clear();
    while (!originStack.empty()) {
        originTemp.push_back(originStack.top());
        originStack.pop();
    }
    while (!targetStack.empty()) {
        targetTemp.push_back(targetStack.top());
        targetStack.pop();
    }
    int originNum = originTemp.size();
    int targetNum = targetTemp.size();

    double minDist;
    int minpos;
    double nowDist;

    for (int i = 0; i < targetNum; i++) {
        int r = RADIUS; 
        if (flag == 1) r = 0;
        minDist = MAX_INT, minpos = -1;

        for (int j = max(0, i - r); j <= min(originNum - 1, i + RADIUS); j++) {
            nowDist = dist(leftTarget[i], rightTarget[i], leftOrigin[j], rightOrigin[j]);
            if (nowDist < ENTROPY_THRESHOLD) {

                if (minDist > nowDist) {
                    minDist = nowDist;
                    minpos = j;
                }
            }
        }
        
        // check inferred_chunks_DB
        if (minpos != -1) {
            leveldb::Status findStatus;
            leveldb::Slice inputKey(targetTemp[i].key, FP_SIZE);
            std::string getValue;
            findStatus = inferred_chunks_DB->Get(leveldb::ReadOptions(), inputKey, &getValue);

            //-------------------add test---------------------
            // print_fp(originTemp[minpos]);print_fp(targetTemp[i]);printf("\n");
            // printf("the dist of this two is %lf:\n", minDist);
            //-------------------test end---------------------

            if (!findStatus.ok()) {
                if (queue_M_G.size() + 1 > QUEUE_LIMIT_W) break;
                leveldb::Slice valueKey(originTemp[minpos].key, FP_SIZE);
                findStatus = inferred_chunks_DB->Put(leveldb::WriteOptions(), inputKey, valueKey);
                findStatus = inferred_chunks_DB->Put(leveldb::WriteOptions(), valueKey, inputKey);

                leftOrigin[minpos] = MAX_INT;
                rightOrigin[minpos] = MAX_INT;
                queue_M_G.push(originTemp[minpos]);
                queue_C_G.push(targetTemp[i]);
            }
        }
    }
}


/* determine whether chunkhash is equal & Encryption_ATTACK */
int equal(char *ptext, char *ctext)
{
    if (!E_ATTACK) {
        if (memcmp(ptext, ctext, FP_SIZE) == 0) {
            return 1;
        }
        else
            return 0;
    } else {

        AES_KEY aes;
        if (AES_set_encrypt_key((unsigned char*)aeskey, 128, &aes) < 0)
        {
            printf("AES_set_decrypt_key failed\n");
            return 0;
        }
        
        char current_hash[16];
        char encrypt_hash[16];

        memset(current_hash, 0, 16);
        memcpy(current_hash, ptext , FP_SIZE);
        memcpy(current_hash + FP_SIZE, ptext , FP_SIZE);
        memcpy(current_hash + 2 * FP_SIZE, ptext , 4);

        memset(encrypt_hash, 0, 16);
        AES_encrypt((unsigned char *)current_hash, (unsigned char *)encrypt_hash, &aes);
        
        if (memcmp(encrypt_hash + 1, ctext, FP_SIZE) == 0)
            return 1;
        else {
            return 0;
        }
    }
}

void main_loop()
{
    ansq.clear();

    stack<node> tmp;
    stack<node> tmp_origin_M;
    stack<node> tmp_target_C;
    stack<node> omp;

    double origin_left_entropy_M[MAX_UO];
    double target_left_entropy_C[MAX_UT];
    double origin_right_entropy_M[MAX_UO];
    double target_right_entropy_C[MAX_UT];

    if (LEAKAGE_RATE == 0) {

        db_insert(origin_M, TOP_FREQUENT_U);
        // inverting chunk sequence (i.e., sort u-frequent chunks by frequency)
        while (!frequency_queue_M.empty()) {
            tmp_origin_M.push(frequency_queue_M.top());
            frequency_queue_M.pop();
        }

        db_insert(target_C, TOP_FREQUENT_U);
        // inverting the sequence (i.e., sort u-frequent chunks by frequency)
        while (!frequency_queue_M.empty()) {
            tmp_target_C.push(frequency_queue_M.top());
            frequency_queue_M.pop();
        }

        // calculate the top-u frequent chunks' entropy
        get_entropy(tmp_origin_M, origin_left_entropy_M, origin_right_entropy_M, 0, 5000);
        get_entropy(tmp_target_C, target_left_entropy_C, target_right_entropy_C, 1, 5000);

        // infer plaintext-ciphertext pairs
        peakToPair(origin_left_entropy_M, origin_right_entropy_M, target_left_entropy_C, target_right_entropy_C,
                   tmp_origin_M, tmp_target_C, 0);

        clear(tmp_origin_M);
        clear(tmp_target_C);
    }

    // MAIN LOOP
    while (!queue_M_G.empty() && !queue_C_G.empty()) {
        // E_attck check
        if (equal(queue_M_G.front().key, queue_C_G.front().key)) {
            ansq.push_back(queue_M_G.front());
            if (queue_C_G.front().count < 50) smallChunkAmount++;
            correct++;
        } else {
            //-----------add test here-----------------------
            //		printf("error pair:");
            //		print_fp(queue_M_G.front());print_fp(queue_C_G.front());
            //		printf("\n");
            if (queue_M_G.front().count > 50)
                bigChunkError++;
            else
                smallChunkError++;
            //-----------test end ---------------------------
        }

        // clear
        clear(frequency_queue_M);
        clear(frequency_queue_C);
        clear(omp);
        clear(tmp);

        // Iteration paradigm
        neighbor_DB_insert(LEFT_NEIGHBOR_FLAG, 0, queue_M_G.front().key, NEIGHBOR_TOP_FREQUENT_V);
        neighbor_DB_insert(LEFT_NEIGHBOR_FLAG, 1, queue_C_G.front().key, NEIGHBOR_TOP_FREQUENT_V);

        // sort chunks by frequency
        while (!frequency_queue_M.empty() && !frequency_queue_C.empty()) {
            omp.push(frequency_queue_M.top());
            tmp.push(frequency_queue_C.top());
            frequency_queue_M.pop();
            frequency_queue_C.pop();
        }
        clear(frequency_queue_M);
        clear(frequency_queue_C);

        get_entropy(omp, origin_left_entropy_M, origin_right_entropy_M, 0, 5000);
        get_entropy(tmp, target_left_entropy_C, target_right_entropy_C, 1, 5000);

        peakToPair(origin_left_entropy_M, origin_right_entropy_M, target_left_entropy_C, 
                   target_right_entropy_C, omp,tmp, 1);

        // clear
        clear(frequency_queue_M);
        clear(frequency_queue_C);
        clear(omp);
        clear(tmp);

        neighbor_DB_insert(RIGHT_NEIGHBOR_FLAG, 0, queue_M_G.front().key, NEIGHBOR_TOP_FREQUENT_V);
        neighbor_DB_insert(RIGHT_NEIGHBOR_FLAG, 1, queue_C_G.front().key, NEIGHBOR_TOP_FREQUENT_V);

        while (!frequency_queue_M.empty() && !frequency_queue_C.empty()) {
            omp.push(frequency_queue_M.top());
            tmp.push(frequency_queue_C.top());
            frequency_queue_M.pop();
            frequency_queue_C.pop();
        }

        clear(frequency_queue_M);
        clear(frequency_queue_C);

        get_entropy(omp, origin_left_entropy_M, origin_right_entropy_M, 0, 5000);
        get_entropy(tmp, target_left_entropy_C, target_right_entropy_C, 1, 5000);

        peakToPair(origin_left_entropy_M, origin_right_entropy_M, target_left_entropy_C,
                   target_right_entropy_C, omp,tmp, 1);
        
        clear(frequency_queue_M);
        clear(frequency_queue_C);
        clear(omp);
        clear(tmp);

        queue_M_G.pop();
        queue_C_G.pop();
        involve++;
    }
    printf("Leaked chunks:%ld\nLeaked chunks appearing in auxiliary information: %ld\n", leaked_common, leaked_chunks);

}

int main(int argc, char *argv[])
{
    init_db(argv[6], &origin_M); // refer to original F_db
    init_db(argv[7], &left_M); // refer to original L_db
    init_db(argv[8], &right_M); // refer to original R_db

    init_db(argv[9], &target_C); // refer to target F_db
    init_db(argv[10], &left_C); // refer to target_C L_db
    init_db(argv[11], &right_C); // refer to target_C R_db

    char inference_db_path[20] = "./inference-db/";
    init_db(inference_db_path, &inferred_chunks_DB); // store all unique for inferring chunks

    char ground_truth_db_path[20] = "./ground-truth/";
    init_db(ground_truth_db_path, &ground_truth_DB); // refer to ground-truth db, using for simulating attack

    TOP_FREQUENT_U = stoi(argv[1]); // u->top u frequency 
    RADIUS = stoi(argv[2]); // r->radius 
    ENTROPY_THRESHOLD = stof(argv[3]); // t->threshold
    NEIGHBOR_TOP_FREQUENT_V = stoi(argv[4]);
    LEAKAGE_RATE = 0;
    
    printf("%llu %llu %lf %llu %lf\n",TOP_FREQUENT_U,RADIUS,ENTROPY_THRESHOLD,NEIGHBOR_TOP_FREQUENT_V,LEAKAGE_RATE);
    QUEUE_LIMIT_W = W; // w-> maximum of inferred pair

    enqueue_leaked_chunks();
    main_loop();
    printf("\nSuccessfully inferred following chunks:\n");
    /*
    printf("\nSuccessfully inferred following chunks:\n");
    while(!ansq.empty()) {
        node tmp = ansq.back();
        printf("%.2hhx",tmp.key[0]);
        for (int i = 1;i < FP_SIZE; i++)
                printf(":%.2hhx", tmp.key[i]);
        printf("\n");
        ansq.pop_back();
    }
    */
    printf("Involve:%ld\n", involve);
    printf("Correct:%ld\n", correct + leaked_common - leaked_chunks);

    printf("Inferred rate: %lf%%\n", (double)((double)(correct) / total) * 100.0);
    printf("Precision: %lf%%\n", (double)correct/involve*100.0);

    printf("Small chunk:%ld\n", smallChunkAmount);
    printf("Big chunk error:%ld\nSmall error:%ld\n", bigChunkError, smallChunkError);
    delete origin_M;
    delete left_M;
    delete right_M;

    delete target_C;
    delete left_C;
    delete right_C;
    return 0;
}
