#include <assert.h>
#include <stdio.h>
#include <queue>
#include <stack>
#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <deque>
#include <openssl/bn.h>
#include <openssl/sha.h>
#include <inttypes.h>
#include "leveldb/db.h"
#include <map>
#include <cmath>
using namespace std;

#define FP_SIZE 6

uint64_t TH_K;
uint64_t INIT;
uint64_t QUEUE_LIMIT;
double LEAK_RATE;

struct node
{
	char key[FP_SIZE];
	uint64_t count;
};

struct cmp
{
	bool operator()(node a, node b) 
	{
		return a.count > b.count;
	}
};

leveldb::DB *origin;	// F_M
leveldb::DB *left_o;	// L_M
leveldb::DB *right_o;	// R_M

leveldb::DB *target;	// F_C
leveldb::DB *left_t;	// L_C
leveldb::DB *right_t;	// R_C
leveldb::DB *size_db;
// unique db is used to record all inferred chunks
leveldb::DB *uniq;

priority_queue<node, vector<node>, cmp > pq;
priority_queue<node, vector<node>, cmp > pc;

// q_o and q_t implement inferred set G
queue<node> q_o;
queue<node> q_t;

map<uint64_t, priority_queue<node, vector<node>, cmp > >bucket_o;
map<uint64_t, priority_queue<node, vector<node>, cmp > >bucket_t;


uint64_t total = 0;
uint64_t common = 0;
uint64_t correct = 0;
uint64_t uniq_count = 0;
uint64_t involve = 0;
uint64_t leak = 0;

void init_db(std::string db_name, int type) 
{
	leveldb::DB *db;
	leveldb::Options options;
	options.create_if_missing = true;
	leveldb::Status status = leveldb::DB::Open(options, db_name.c_str(), &db);
	assert(status.ok());
	assert(db != NULL);

	if(type == 1) origin = db;
	if(type == 11) left_o = db;
	if(type == 12) right_o = db;
	if(type == 2) target = db;
	if(type == 21) left_t = db;
	if(type == 22) right_t = db;
	if(type == 3) uniq = db;
	if(type == 4) size_db = db;
}

// enqueue leaked chunks based on leakage rate
void stat_db()
{
	leveldb::Iterator* it = target->NewIterator(leveldb::ReadOptions());
	leveldb::Status status;
	std::string existing_value;

	for (it->SeekToFirst(); it->Valid(); it->Next())
	{
		total ++;
		status = origin->Get(leveldb::ReadOptions(), it->key(), &existing_value);
		if(LEAK_RATE != 0 && rand()%10000 <= LEAK_RATE*10000)
		{
			common ++;
			if (status.ok())
			{
				node entry_o;
				memcpy(entry_o.key, it->key().ToString().c_str(), FP_SIZE);
			//	entry_o.count = strtoimax(existing_value.data(), NULL, 10);
				entry_o.count = *(uint64_t *)existing_value.c_str();
				q_o.push(entry_o);

				node entry_t;
				memcpy(entry_t.key, it->key().ToString().c_str(), FP_SIZE);
			//	entry_t.count = strtoimax(it->value().ToString().c_str(), NULL, 10);
				entry_t.count = *(uint64_t *)it->value().ToString().c_str();
				leveldb::Status s;
				leveldb::Slice k(entry_t.key, FP_SIZE);
				char buf[32];
				memset(buf, 0, 32);
				sprintf(buf, "%lu", entry_t.count);
				leveldb::Slice u(buf, sizeof(uint64_t));
				s = uniq->Put(leveldb::WriteOptions(), k, u);
				q_t.push(entry_t);

				leak ++;		
			}else
			if(!status.IsNotFound())
			{printf("%s\n", status.ToString().c_str());exit(1);}//leveldb error!
		}
	}

	printf("Total number of unique ciphertext chunks: %lu\nLeakage rate: %lf%%\n", total,(double)(LEAK_RATE * 100.0));
}

void print_fp(node a)
{
	int len = 0;
	for (len = 0; len < FP_SIZE; len++)
	{
		printf("%02x", (unsigned char)a.key[len]);
		if (len < FP_SIZE - 1) printf(":");
	}
	printf("\t");
}

void left_insert(int type, char* fp, uint64_t k)
{
	leveldb::Status status;
	leveldb::Slice key(fp, FP_SIZE);
	std::string existing_value;
	priority_queue<node, vector<node>, cmp >* pt;

	//if(type == 0) pt = &pq;
	//else pt = &pc;


	uint64_t len = 0;
	char tar[FP_SIZE];
	uint64_t tmp;

	if(type == 0) status = left_o->Get(leveldb::ReadOptions(), key, &existing_value);
	else 	status = left_t->Get(leveldb::ReadOptions(), key, &existing_value);
	//printf("%s/n",status.ToString().c_str());
	if(status.ok())
	{
		//printf("%d\n", existing_value.size());
		while(len < existing_value.size())
		{
			memcpy(tar, existing_value.c_str()+len, FP_SIZE);

			const char* t_int = existing_value.c_str()+len+FP_SIZE;
			tmp = *(uint64_t*)t_int;
			node entry;
			memcpy(entry.key, tar, FP_SIZE);
			entry.count = tmp;
			//insert bucket here.
			std::string sizer;
			leveldb::Slice sizekey(entry.key, FP_SIZE);
			status = size_db->Get(leveldb::ReadOptions(), sizekey, &sizer);
			uint64_t intsize2 =  *(uint64_t*)sizer.c_str();
			uint64_t intsize = ceil((double)intsize2/16.0);
			//printf("intsize:%lld\n",intsize2);
			if(type == 0) pt = &bucket_o[intsize];
			else pt = &bucket_t[intsize];
			if(pt->size()<k)
			{
				pt->push(entry);
			}else
			{
				node min = pt->top();
				if(tmp > min.count)
				{
					pt->pop();
					pt->push(entry);
				}
			}

			//printf("%d\n", len);
			len += (FP_SIZE+sizeof(uint64_t));	
		}
	}else
	if(!status.IsNotFound()){printf("%s\n", status.ToString().c_str());exit(1);}//leveldb error!
}

void right_insert(int type, char* fp, uint64_t k)
{
	leveldb::Status status;
	leveldb::Slice key(fp, FP_SIZE);
	std::string existing_value;
	priority_queue<node, vector<node>, cmp >* pt;

	//if(type == 0) pt = &pq;
	//else pt = &pc;


	uint64_t len = 0;
	char tar[FP_SIZE];
	uint64_t tmp;

	if(type == 0)status = right_o->Get(leveldb::ReadOptions(), key, &existing_value);
	else status = right_t->Get(leveldb::ReadOptions(), key, &existing_value);

	if(status.ok())
	{
		while(len < existing_value.size())
		{
			memcpy(tar, existing_value.c_str()+len, FP_SIZE);
			const char* t_int = existing_value.c_str()+len+FP_SIZE;
			tmp = *(uint64_t*)t_int;
			node entry;
			//printf("1\n");
			memcpy(entry.key, tar, FP_SIZE);
			entry.count = tmp;
			//insert bucket here.
                        std::string sizer;
                        leveldb::Slice sizekey(entry.key, FP_SIZE);
                        status = size_db->Get(leveldb::ReadOptions(), sizekey, &sizer);
			//printf("2\n");
                        uint64_t intsize2 =  *(uint64_t*)sizer.c_str();
			uint64_t intsize = ceil((double)intsize2/16.0);
                        if(type == 0) pt = &bucket_o[intsize];
                        else pt = &bucket_t[intsize];
			//printf("3\n");
			if(pt->size()<k)
			{
				pt->push(entry);
			}else
			{
				node min = pt->top();
				if(tmp > min.count)
				{
					pt->pop();
					pt->push(entry);
				}
			}


			len += (FP_SIZE+sizeof(uint64_t));	
		}
	}else
	if(!status.IsNotFound()){printf("%s\n", status.ToString().c_str());exit(1);}

}

// insert top-k frequent chunks (in db) into pq
void db_insert(leveldb::DB* db, uint64_t k, int type)
{
	leveldb::Status status;
	priority_queue<node, vector<node>, cmp >* pt;
	leveldb::Iterator* it = db->NewIterator(leveldb::ReadOptions());
	for(it->SeekToFirst(); it->Valid(); it->Next())
	{
		//uint64_t val = strtoimax(it->value().ToString().c_str(), NULL, 10);
		uint64_t val = *(uint64_t *)it->value().ToString().c_str();
		node entry;
		memcpy(entry.key, it->key().ToString().c_str(), FP_SIZE);
		leveldb::Slice sizekey(entry.key, FP_SIZE);
		std::string sizer = "";
                status = size_db->Get(leveldb::ReadOptions(), sizekey, &sizer);
                   //printf("2\n") 
                uint64_t intsize2 =  *(uint64_t*)sizer.c_str();
                uint64_t intsize = ceil((double)intsize2/16.0);
                
		if(type == 0) pt = &bucket_o[intsize];
                else pt = &bucket_t[intsize];

		entry.count = val;
		if(pt -> size() < k) 
		{
			pt->push(entry);
		}else
		{
			node min = pt->top();
			if(val > min.count)
			{
				pt->pop();
				pt->push(entry);
			}
		}
	}	

}
vector <node>ansq;
void main_loop()
{
	ansq.clear();
	stack<node> tmp;
	stack<node> omp;
	if(LEAK_RATE == 0)
	{
	
		db_insert(origin, INIT, 0);
		db_insert(target, INIT, 1);
		std::map<uint64_t, priority_queue<node, vector<node>, cmp > >::iterator tit;
		for(tit = bucket_t.begin(); tit!= bucket_t.end(); tit++)
                {
			uint64_t nowsize = tit->first;
                	priority_queue<node, vector<node>, cmp > *tpt, *opt;
                	tpt = &tit->second;
                	opt = &bucket_o[nowsize];
                	while (!(*opt).empty() && !(*tpt).empty())
                	{
                        	omp.push((*opt).top());
                        	tmp.push((*tpt).top());
                        	(*opt).pop();
                        	(*tpt).pop();
                	}
			
			
		}
		while (!tmp.empty() && !omp.empty())
                {
                        leveldb::Status ss;
                        leveldb::Slice kk(tmp.top().key, FP_SIZE);
                        std::string ex;
                        ss = uniq->Get(leveldb::ReadOptions(), kk, &ex);

                        // if  ciphertext chunk has not been inferred
                        if(!ss.ok())
                        {
                                if(q_o.size()+1 > QUEUE_LIMIT) break;
                                char buf[32];
                                memset(buf,0,32);
                                sprintf(buf, "%lu", tmp.top().count);
                                leveldb::Slice uu(buf, sizeof(uint64_t));
                                ss = uniq->Put(leveldb::WriteOptions(), kk, uu);
                                q_o.push(omp.top());
                                q_t.push(tmp.top());
                        }
                        omp.pop();
                        tmp.pop();
                }

	}
	// MAIN LOOP
	while(!q_o.empty() && !q_t.empty())
	{
		if(memcmp(q_o.front().key, q_t.front().key, FP_SIZE) == 0) 
		{
			ansq.push_back(q_o.front());
			correct++;
		}
		std::map<uint64_t, priority_queue<node, vector<node>, cmp > >::iterator cit;
		for(cit = bucket_o.begin(); cit!= bucket_o.end(); cit++)
		{
			if(!cit->second.empty())cit->second.pop();
		}
		for(cit = bucket_t.begin(); cit!= bucket_t.end(); cit++)
                {
                        if(!cit->second.empty())cit->second.pop();
                }
		while(!omp.empty()) omp.pop();
		while(!tmp.empty()) tmp.pop();

		left_insert(0, q_o.front().key, TH_K);
		left_insert(1, q_t.front().key, TH_K);
		
		//printf("12334\n");
		std::map<uint64_t, priority_queue<node, vector<node>, cmp > >::iterator tit;
		for(tit = bucket_t.begin(); tit!= bucket_t.end(); tit++)
		{
		//printf("%d\n",tit->second.size());
		uint64_t nowsize = tit->first;
		priority_queue<node, vector<node>, cmp > *tpt, *opt;
		tpt = &tit->second;
		opt = &bucket_o[nowsize];
		while (!(*opt).empty() && !(*tpt).empty())
		{
			omp.push((*opt).top());
			tmp.push((*tpt).top());
			(*opt).pop();
			(*tpt).pop();
		}
		//printf("%d  %d\n",omp.size(),tmp.size());
		while (!tmp.empty() && !omp.empty())
		{
			leveldb::Status ss;
			leveldb::Slice kk(tmp.top().key, FP_SIZE);
			std::string ex;
			ss = uniq->Get(leveldb::ReadOptions(), kk, &ex);

			// if  ciphertext chunk has not been inferred
			if(!ss.ok())
			{
				if(q_o.size()+1 > QUEUE_LIMIT) break;
				char buf[32];
				memset(buf,0,32);
				sprintf(buf, "%lu", tmp.top().count);
				leveldb::Slice uu(buf, sizeof(uint64_t));
				ss = uniq->Put(leveldb::WriteOptions(), kk, uu);
				q_o.push(omp.top());
				q_t.push(tmp.top());
			}
			omp.pop();
			tmp.pop();
		}
		}
		//while(!pq.empty()) pq.pop();
		//while(!pc.empty()) pc.pop();
		while(!omp.empty()) omp.pop();
		while(!tmp.empty()) tmp.pop();
		
                for(cit = bucket_o.begin(); cit!= bucket_o.end(); cit++)
                {
                        if(!cit->second.empty())cit->second.pop();
                }
                for(cit = bucket_t.begin(); cit!= bucket_t.end(); cit++)
                {
                        if(!cit->second.empty())cit->second.pop();
                }

	//	bucket_o.clear();
          //      bucket_t.clear();

		right_insert(0, q_o.front().key, TH_K);
		//printf("12335\n");
		right_insert(1, q_t.front().key, TH_K);
		//printf("12336\n");
		for(tit = bucket_t.begin(); tit!= bucket_t.end(); tit++)
		{
		uint64_t nowsize = tit->first;
                priority_queue<node, vector<node>, cmp > *tpt, *opt;
                tpt = &tit->second;
                opt = &bucket_o[nowsize];
		//printf("12336\n");
		while (!(*opt).empty() && !(*tpt).empty())
		{
			//omp.push(pq.top());
			//tmp.push(pc.top());
			omp.push((*opt).top());
                        tmp.push((*tpt).top());
			(*opt).pop();
			(*tpt).pop();
		}
		//printf("12337\n");
		while (!tmp.empty() && !omp.empty())
		{
			leveldb::Status ss;
			leveldb::Slice kk(tmp.top().key, FP_SIZE);
			std::string ex;
			ss = uniq->Get(leveldb::ReadOptions(), kk, &ex);

			if(!ss.ok())
			{
				if(q_o.size()+1 > QUEUE_LIMIT) break;
				char buf[32];
				memset(buf,0,32);
				sprintf(buf, "%lu", tmp.top().count);
				leveldb::Slice uu(buf, sizeof(uint64_t));
				ss = uniq->Put(leveldb::WriteOptions(), kk, uu);
				q_o.push(omp.top());
				q_t.push(tmp.top());
			}
			omp.pop();
			tmp.pop();
		}
		//printf("12338\n");
		}
		q_o.pop();
		q_t.pop();
		involve ++;
		//printf("involve:%lld\n", involve);

	}
	// printf("Leaked chunks:%lu\nLeaked chunks appearing in auxiliary information:%lu\n", common, leak);
	printf("involve:%lld\n", involve);
	printf("Correct inferences: %lu\nInference rate: %lf%%\n", correct + common -leak, (double)((double)(correct + common - leak)/total)*100.0);
}

int main (int argc, char *argv[])
{
	init_db(argv[5], 1);// refer to original F_db
	init_db(argv[6], 11);// refer to original L_db
	init_db(argv[7], 12);// refer to original R_db
	init_db(argv[8], 2);// refer to targrt F_db
	init_db(argv[9], 21);// refer to target L_db
	init_db(argv[10], 22);// refer to target R_db
	init_db(argv[11], 4);
	init_db("./inference-db/", 3);
		
	INIT = atoi(argv[1]);	// u
	TH_K = atoi(argv[2]);	// v
	QUEUE_LIMIT = atoi(argv[3]);	// w

	LEAK_RATE = atof(argv[4]);

	stat_db();
	main_loop();
	printf("\nSuccessfully inferred following chunks:\n");
	while(!ansq.empty())
        {
                node tmp = ansq.back();
                printf("%.2hhx",tmp.key[0]);
                for (int i = 1;i < FP_SIZE; i++)
                        printf(":%.2hhx", tmp.key[i]);
                printf("\n");
                ansq.pop_back();
        }
	return 0;
}
